//
//  tank
//
//  Import and Playlist Daemon for autoradio project
//
//
//  Copyright (C) 2017-2018 Christian Pointner <equinox@helsinki.at>
//
//  This file is part of tank.
//
//  tank is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  any later version.
//
//  tank is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with tank. If not, see <http://www.gnu.org/licenses/>.
//

package importer

import (
	"context"
	"io"
	"net/url"
	"sync/atomic"
	"time"
	"unsafe"

	"gitlab.servus.at/autoradio/tank/store"
)

// Job represents a import job for the file identified by Group and ID
type Job struct {
	im             *Importer
	ctx            context.Context
	cancel         context.CancelFunc
	State          JobState    `json:"state"`
	CreatedAt      Timestamp   `json:"created"`
	StartedAt      Timestamp   `json:"started,omitempty"`
	ID             uint64      `json:"id"`
	Group          string      `json:"group"`
	User           string      `json:"user"`
	Source         SourceURL   `json:"source"`
	RefID          string      `json:"ref-id,omitempty"`
	Progress       JobProgress `json:"progress"`
	source         *JobSource
	sourceAttached chan struct{}
}

type Jobs []*Job

func (job *Job) run() (err error) {
	// this function must never-ever hang or we block a worker forever - watch your context kids!!

	// when this function returns the job is either done or aborted...
	// either way job.cancel() needs to be called and the job needs to be removed from the inventory
	defer job.cancel()
	defer job.cleanup()

	job.prepareSource()

	if !atomic.CompareAndSwapUint32((*uint32)(&job.State), uint32(JobPending), uint32(JobRunning)) {
		job.im.store.UpdateFileImportState(job.Group, job.ID, store.ImportAborted, false)
		// the job was canceled before job.Start() could initialize the context and because of this
		// job.Cancel() only set the state to JobCanceled... so we simulate a canceled context here
		return context.Canceled
	}

	job.StartedAt.set(time.Now())
	job.im.dbgLog.Printf("running import for %s/%d from: %s", job.Group, job.ID, job.Source.String())

	job.im.store.UpdateFileImportState(job.Group, job.ID, store.ImportRunning, false)
	if err = job.fetch(); err != nil {
		job.im.store.UpdateFileImportState(job.Group, job.ID, store.ImportAborted, false)
		// send result to all done subscriptions
		return
	}
	if err = job.normalize(); err != nil {
		job.im.store.UpdateFileImportState(job.Group, job.ID, store.ImportAborted, false)
		// send result to all done subscriptions
		return
	}
	job.im.store.UpdateFileImportState(job.Group, job.ID, store.ImportDone, true)
	// send result to all done subscriptions
	return
}

// Start adds job to the work queue of the importer. If timeout is > 0 it will create a context
// WithTimeout with ctx as its parent or a WithCancel otherwise. Of course you may pass on a context
// that already has a deadline or timeout set. In this case you may set timeout to 0. But
// context.Context requires you to cleanup resources with after all work is done by calling it's
// CancelFunc. This is not easy to do from stateless interfaces. In this case use context.Background()
// and set the timeout to a appropiate value. The job will then take care of freeing up the resources.
func (job *Job) Start(ctx context.Context, timeout time.Duration) error {
	if !atomic.CompareAndSwapUint32((*uint32)(&job.State), uint32(JobNew), uint32(JobInitializing)) {
		if atomic.LoadUint32((*uint32)(&job.State)) == uint32(JobCanceled) {
			// the job has already been canceled and thus should just be removed from the inventory
			job.cleanup()
			return ErrAlreadyCanceled
		}
		return nil
	}

	// shall we use a global context here so we can cancel all jobs at once? i.e. for forced shutdown...
	if timeout > 0 {
		job.ctx, job.cancel = context.WithTimeout(ctx, timeout)
	} else {
		job.ctx, job.cancel = context.WithCancel(ctx)
	}
	// from here on we need to take care that job.cancel() will be called in any case so that resources
	// allocated by job.ctx are freed up.

	// although the job is not really enqueued yet we need to update the file state here because after the
	// select there is a race between this update call and the one performed in the beginning of job.run()
	job.im.store.UpdateFileImportState(job.Group, job.ID, store.ImportPending, false)
	select {
	case job.im.work <- job:
		if !atomic.CompareAndSwapUint32((*uint32)(&job.State), uint32(JobInitializing), uint32(JobPending)) {
			// we just posted the job onto the work channel but job.Cancel() has apparently been called while
			// we were setting up the context. Because of this job.Cancel() couldn't use the cancel function
			// provided by the context. There is not much that we can do here... job.run() needs to clean up
			// the rest and remove the job from the inventory.
			return ErrAlreadyCanceled
		}
	default:
		// the work channel is already full so we need to drop all new jobs...
		job.im.store.UpdateFileImportState(job.Group, job.ID, store.ImportAborted, false)
		job.cancel()
		job.cleanup()
		return ErrTooManyJobs
	}

	// the job was successfully enqueued now and awaits a worker to call job.run().
	return nil
}

func (job *Job) AttachSource(length uint64, r io.Reader) error {
	if state := atomic.LoadUint32((*uint32)(&job.State)); state != uint32(JobRunning) {
		return ErrImportNotRunning
	}

	src := &JobSource{length, r}
	if ok := atomic.CompareAndSwapPointer((*unsafe.Pointer)(unsafe.Pointer(&job.source)), unsafe.Pointer(nil), unsafe.Pointer(src)); !ok {
		return ErrSourceAlreadyAttached
	}
	close(job.sourceAttached)
	return nil
}

func (job *Job) Cancel() {
	oldState := atomic.SwapUint32((*uint32)(&job.State), uint32(JobCanceled))
	// this next line is why we need to make sure JobCanceled is smaller than all the other states.
	if oldState < uint32(JobPending) {
		// job.cancel() might not yet be initialized yet, setting the state to canceled is the
		// only thing we can do for now... job.Start() or job.run() will take care of the rest.
		return
	}
	job.cancel()
}

// TODO: handle subscriptions to done

func (job *Job) cleanup() {
	atomic.StoreUint32((*uint32)(&job.State), uint32(JobDestroying))

	// we are not really interested in the result of the deleteJob here,
	// the only possible error would be ErrNotFound anyway... no need to wait for that.
	go job.im.deleteJob(job.Group, job.ID)
}

func newJob(im *Importer, group string, id uint64, src url.URL, user, refID string) *Job {
	job := &Job{im: im, Group: group, ID: id, Source: SourceURL(src), User: user, RefID: refID}
	job.State = JobNew
	job.CreatedAt.set(time.Now())
	job.sourceAttached = make(chan struct{})
	return job
}
